
/*
  Copyright (C) 2002-2008,2012  Dmitry V. Levin <ldv@altlinux.org>

  CheckIgnore method for packageof.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "chkign.h"
#include "cwd.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <dirent.h>
#include <sys/stat.h>

static void __attribute__ ((noreturn))
reg_fatal(int rc, regex_t * handle)
{
	size_t  len = regerror(rc, handle, 0, 0);
	char   *errbuf = (char *) alloca(len);

	regerror(rc, handle, errbuf, len);
	error(EXIT_FAILURE, 0, "regcomp: %s", errbuf);
	exit(EXIT_FAILURE);
}

CheckIgnore::~CheckIgnore()
{
	if (!handle)
		return;

	regfree(handle);
	delete handle;
}

int
CheckIgnore::add(const char *fn)
{
	int     fd = open(fn, O_RDONLY);

	if (fd < 0)
		return 0;

	struct stat st;

	if (fstat(fd, &st) < 0)
		return 0;

	if (S_ISDIR(st.st_mode))
	{
		// directory processing.
		close(fd);
		CWD     cwd;

		if (chdir(fn) < 0)
			return 0;

		DIR    *dir = opendir(".");

		if (!dir)
			return 0;

		struct dirent *ent;

		while ((ent = readdir(dir)))
		{
			const char *name = ent->d_name;

			if (!name[0] || name[0] == '.')
				continue;

			if (stat(name, &st))
				continue;

			if (!S_ISREG(st.st_mode))
				continue;

			string  s(name);

			ign_list[s] = 1;

			if (st.st_blocks)
				add(name);
		}

		closedir(dir);
		return 1;
	}
	// regular file processing.
	FILE   *fp = fdopen(fd, "r");

	if (!fp)
		return 0;

	int     rc = 1;
	char    line[PATH_MAX];

	while (fgets(line, sizeof(line), fp))
	{
		char   *p;

		for (p = line; *p; ++p)
			if (!isspace(*p))
				break;

		for (size_t i = strlen(p) - 1; i > 0; --i)
		{
			if (isspace(p[i]))
				p[i] = '\0';
			else
				break;
		}

		if (!*p || *p == '#')
			continue;

		string  s(p);

		ign_list[s] = 1;
	}

	if (!feof(fp))
		rc = 0;

	if (fclose(fp))
		rc = 0;

	return rc;
}

void
CheckIgnore::compile()
{
	if (handle)
		return;

	string  pattern;

	for (PkgMapType::const_iterator i = ign_list.begin();
	     i != ign_list.end(); i++)
	{
		if (pattern.length())
			pattern += "|^";
		else
			pattern = '^';
		pattern += i->first;
		pattern += '$';
	}

	/* No need to keep ign_list anymore. */
	ign_list.clear();

	const char *p = pattern.c_str();

#ifdef DEBUG
	error(0, 0, "compile: <pattern=%s>", p);
#endif

	if (!*p)
		return;

	handle = new regex_t;

	int     rc = regcomp(handle, p, REG_EXTENDED | REG_NOSUB);

	if (rc)
		reg_fatal(rc, handle);
}

int
CheckIgnore::is_ignore(const string &str)
{
	if (!handle)
		return 0;

	/* Check cache first. */
	int    &val = ign_cache[str];

	if (val > 0)
		return 1;
	else if (val < 0)
		return 0;

	/* Not found in cache, run regexec. */
	const char *sample = str.c_str();
	int     rc = regexec(handle, sample, 0, 0, 0);

	if (!rc)
	{
		val = 1;
		return 1;
	} else if (REG_NOMATCH == rc)
	{
		val = -1;
		return 0;
	}

	reg_fatal(rc, handle);
}
