/*
  Copyright (C) 2000-2008  Dmitry V. Levin <ldv@altlinux.org>

  Generates list of duplicated packages.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <cstdio>
#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <stdint.h>
#include <error.h>
#include <unistd.h>
#include <sys/stat.h>
/* due to rpmEVRDTCompare() */
#define ALT_RPM_API
#include <rpmlib.h>

#include <string>
#include <set>
#include <map>
#include <algorithm>
#include <functional>
#include <iostream> 

using std::string;

static bool is_reverse;

static char *
xstrdup(const char *s)
{
	char   *r = strdup(s);

	if (!r)
		error(EXIT_FAILURE, errno, "strdup");
	return r;
}

struct fpkginfo
{
	dev_t device;
	ino_t inode;
	const char *fname;
	rpmEVRDT verinfo;
};

struct fpkginfo_less:
	public std::binary_function < const fpkginfo &, const fpkginfo &, bool >
{
	bool    operator  () (const fpkginfo &a, const fpkginfo &b) const
	{
		if (a.device == b.device && a.inode == b.inode)
			return false;
		return rpmEVRDTCompare(&a.verinfo, &b.verinfo) < 0;
	}
};

typedef std::set < fpkginfo, fpkginfo_less > fpkginfo_set;
typedef std::map < string, fpkginfo_set > fpkginfo_set_map;

class print_pkginfo:
	public std::unary_function < fpkginfo_set::value_type, void >
{
	bool started;

    public:
	print_pkginfo(): started(false) {}

	void    operator() (const fpkginfo_set::value_type &info)
	{
		if (started)
			putchar(' ');
		else
			started = true;

		fputs(info.fname, stdout);
	}
};

struct print_set: public std::unary_function < fpkginfo_set_map::value_type, void >
{
	print_set() {}

	void operator() (const fpkginfo_set_map::value_type &pair)
	{
		const fpkginfo_set &set = pair.second;

		if (set.size() <= 1)
			return;
		if (is_reverse)
			for_each(set.rbegin(), set.rend(), print_pkginfo());
		else
			for_each(set.begin(), set.end(), print_pkginfo());
		putchar('\n');
	}
};

static int
add_file (fpkginfo_set_map &map, const char *fname)
{
	FD_t    fd = Fopen(fname, "r.ufdio");

	if (!fd || Ferror(fd))
	{
		error(0, 0, "open of %s failed: %s", fname, Fstrerror(fd));
		if (fd)
			Fclose(fd);
		return EXIT_FAILURE;
	}

	struct stat st;

	if (fstat(Fileno(fd), &st) < 0)
	{
		error(0, errno, "stat of %s failed", fname);
		Fclose(fd);
		return EXIT_FAILURE;
	}

	Header h;
	rpmRC rpmrc = rpmReadPackageFile(NULL, fd, fname, &h);

	Fclose(fd);

	if (rpmrc != RPMRC_OK || h == NULL)
	{
		error(0, 0, "query of %s failed", fname);
		return EXIT_FAILURE;
	}

	const char *name = headerGetString(h, RPMTAG_NAME);
	if (!name)
	{
		error(0, 0, "query of %s failed: no NAME tag found", fname);
		h = headerFree(h);
		return EXIT_FAILURE;
	}

	uint64_t epoch = headerGetNumber(h, RPMTAG_EPOCH);

	const char *version = headerGetAsString(h, RPMTAG_VERSION);
	if (!version)
	{
		error(0, 0,
		      "query of %s failed: no VERSION tag found", fname);
		h = headerFree(h);
		return EXIT_FAILURE;
	}

	const char *release = headerGetAsString(h, RPMTAG_RELEASE);
	if (!release)
	{
		error(0, 0,
		      "query of %s failed: no RELEASE tag found", fname);
		h = headerFree(h);
		return EXIT_FAILURE;
	}

	const char *disttag = headerGetAsString(h, RPMTAG_DISTTAG);

	uint64_t buildtime = headerGetNumber(h, RPMTAG_BUILDTIME);

	const char *arch;

	if (headerIsSource(h))
		arch = "";
	else
	{
		arch = headerGetString(h, RPMTAG_ARCH);
		if (!arch)
		{
			error(0, 0,
			      "query of %s failed: no ARCH tag found", fname);
			h = headerFree(h);
			return EXIT_FAILURE;
		}
	}

	fpkginfo pkg;

	pkg.device = st.st_dev;
	pkg.inode = st.st_ino;
	pkg.fname = xstrdup(fname);
	/* If it was absent, headerGetNumber() returned 0. But that's not important. */
	pkg.verinfo.has_epoch = true;
	pkg.verinfo.epoch = epoch;
	pkg.verinfo.version = version;
	pkg.verinfo.release = release;
	pkg.verinfo.disttag = disttag;
	/* If it was absent, headerGetNumber() returned 0. But that's not important. */
	pkg.verinfo.has_buildtime = true;
	pkg.verinfo.buildtime = buildtime;
	string  mapname(arch);

	mapname += '/';
	mapname += name;

	map[mapname].insert(pkg);

	h = headerFree(h);
	return EXIT_SUCCESS;
}

int
main(int ac, const char *av[])
{
	bool    read_stdin = false;

	if (ac == 2 && av[1][0] == '-' && av[1][1] == '\0')
		read_stdin = true;
	else if (ac <= 1)
		return EXIT_SUCCESS;

	if (rpmReadConfigFiles(0, 0))
		error(EXIT_FAILURE, errno, "rpmReadConfigFiles failed");

	int     status, rc = EXIT_SUCCESS;

	fpkginfo_set_map *valgrind_map = new fpkginfo_set_map;
	static  fpkginfo_set_map & map = *valgrind_map;

	if (read_stdin)
	{
		string  s;

		while (getline(std::cin, s))
		{
			if (s.empty())
				continue;
			if ((status =
			     add_file(map, s.c_str())) != EXIT_SUCCESS)
				rc = status;
		}
	} else
	{
		for (int i = 1; i < ac; ++i)
			if ((status = add_file(map, av[i])) != EXIT_SUCCESS)
				rc = status;
	}

	is_reverse = !strcmp(program_invocation_short_name, "rpmrdups");
	for_each(map.begin(), map.end(), print_set());

	return rc;
}
